/* $Id: blockop.c,v 1.4 2004/01/16 20:38:04 chris Exp $
 *
 * Olivetti M20 PCOS diskette access utility: disk block operations
 * Copyright (C) 2000,2004 by Christian Groessler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "cpgtypes.h"
#include "blockop.h"

int read_file_blocks(FILE *disk,int blockno,int numblks,cpg_u8_t *buffer)
{
    int err;
    while(numblks--) {
        if ((err = read_file_block(disk,blockno,buffer))) {
            fprintf(stderr,"read_file_blocks: error %d\n",err);
            exit(55);
        }
        blockno++;
        buffer += 256;
    }
    return(0);
}

int read_phys_blocks(FILE *disk,int blockno,int numblks,cpg_u8_t *buffer)
{
    int err;
    while(numblks--) {
        if ((err = read_phys_block(disk,blockno,buffer))) {
            fprintf(stderr,"read_phys_blocks: error %d\n",err);
            exit(55);
        }
        blockno++;
        buffer += 256;
    }
    return(0);
}

int write_file_blocks(FILE *disk,int blockno,int numblks,cpg_u8_t *buffer)
{
    int err;
    while(numblks--) {
        if ((err = write_file_block(disk,blockno,buffer))) {
            fprintf(stderr,"write_file_blocks: error %d\n",err);
            exit(65);
        }
        blockno++;
        buffer += 256;
    }
    return(0);
}

int write_phys_blocks(FILE *disk,int blockno,int numblks,cpg_u8_t *buffer)
{
    int err;
    while(numblks--) {
        if ((err = write_phys_block(disk,blockno,buffer))) {
            fprintf(stderr,"write_phys_blocks: error %d\n",err);
            exit(65);
        }
        blockno++;
        buffer += 256;
    }
    return(0);
}

int read_file_block(FILE *disk,int blockno,cpg_u8_t *buffer)
{
    if (blockno >= 0x200) blockno += 0x10;
    return(read_phys_block(disk,blockno,buffer));
}

int write_file_block(FILE *disk,int blockno,cpg_u8_t *buffer)
{
    if (blockno >= 0x200) blockno += 0x10;
    return(write_phys_block(disk,blockno,buffer));
}

int read_phys_block(FILE *disk,int blockno,cpg_u8_t *buffer)
{
    if (blockno > 0x460) {
        fprintf(stderr,"read_phys_block: access outa bounds - block %d\n",blockno);
        exit(50);
    }
    if (fseek(disk,blockno * 256,SEEK_SET)) {
        fprintf(stderr,"read_phys_block: fseek failed: %s\n",strerror(errno));
        exit(51);
    }
    if (fread(buffer,256,1,disk) != 1) {
        fprintf(stderr,"read_phys_block: fread failed %s\n",strerror(errno));
        exit(52);
    }
    return(0);
}

int write_phys_block(FILE *disk,int blockno,cpg_u8_t *buffer)
{
    if (blockno > 0x460) {
        fprintf(stderr,"write_phys_block: access outa bounds - block %d\n",blockno);
        exit(60);
    }
    if (fseek(disk,blockno * 256,SEEK_SET)) {
        fprintf(stderr,"write_phys_block: fseek failed: %s\n",strerror(errno));
        exit(61);
    }
    if (fwrite(buffer,256,1,disk) != 1) {
        fprintf(stderr,"write_phys_block: fwrite failed %s\n",strerror(errno));
        exit(62);
    }
    return(0);
}

/* Local Variables: */
/* c-file-style: "cpg" */
/* c-basic-offset: 4 */
/* End: */
